--
-- Inny's Raycaster
-- Written in November, 2012.
-- Licensed under the WTFPL, see: http://en.wikipedia.org/wiki/WTFPL
--
-- Port of Lode Vandevenne's Raycaster (Copyright 2004-2007)
-- See: http://lodev.org/cgtutor/raycasting.html
-- Original Source Code and License available there.
--

NULLFUNC = function() end
log = NULLFUNC

local quarterPi = math.pi/4
local halfPi = math.pi/2
local pi = math.pi
local twoPi = 2*math.pi

function abs(x) return x < 0 and -x or x end

-- Grid class, 2d array provides convinient set and get vars
do
  Grid = {}
  local MT = { __index = Grid }
  function Grid.new(w, h, d, grid)
    return setmetatable({w=w, h=h, default=d, grid=grid or {}}, MT)
  end

  function Grid:set(x, y, v)
    x, y = math.floor(x), math.floor(y)
    if x > 0 and y > 0 and x < self.w and y < self.h then
      local row = self.grid[y]
      if not row then
        row = {}
        self.grid[y] = row
      end
      row[x] = v
    end
    return self
  end

  local NULLROW = {}
  function Grid:get(x, y)
    x, y = math.floor(x), math.floor(y)
    local row = self.grid[y] or NULLROW
    return row[x] or self.default
  end
end

-- Texture class, loads a texture as 1px wide stripes.
do
  Texture = {}
  local MT = { __index = Texture }
  function Texture.new(filename)
    local self = setmetatable({}, MT)
    self:create(filename)
    return self
  end

  function Texture:create(name)
    self.image = love.graphics.newImage(name)
    self.image:setFilter("nearest", "nearest")
    local sw, sh = self.image:getWidth(), self.image:getHeight()
    self.quads = {}
    local i=1
    for y = 0, sh-1, 16 do
      for x = 0, sw-1, 16 do
        for strip = 0, 15 do
          self.quads[i] = love.graphics.newQuad(x+strip, y, 1, 16, sw, sh)
          i=i+1
        end
      end
    end
  end

  function Texture:draw(column, height, id, offset, z)
    local strip = self.quads[1+((id-1)*16)+math.floor(offset*16)]
    if not strip then return end
    local y = z-(height*0.5)
    local scale = height*0.0625
    love.graphics.drawq(self.image, strip, column, y, 0, 1, scale)
  end
end

-- Raycaster, performs distance checks for each vertical scanline
do
  Raycaster = {}
  local MT = { __index = Raycaster }
  function Raycaster.new(x, y, angle)
    return setmetatable({
      x=x or 0,
      y=y or 0,
      angle=angle or 0,
      fov=math.pi/4,
      scanlines=320,
      planeZ=120,
      dist={},
    }, MT)
  end

  local hpi = math.pi/2
  function Raycaster:recomputeVectors()
    local d = self.angle
    self.dirX, self.dirY = math.cos(d), math.sin(d)
    self.planeX, self.planeY = math.cos(d+hpi)*self.fov, math.sin(d+hpi)*self.fov
    return self
  end

  function Raycaster:setGrid(grid, walls)
    self.grid, self.walls = grid, walls
    return self
  end

  function Raycaster:setEntities(entities, sprites)
    self.entities, self.sprites = entities, sprites
    return self
  end

  function Raycaster:update()
    self:recomputeVectors()
  end

  function Raycaster:draw()
    local z = self.planeZ
    love.graphics.setColor(0, 128, 255)
    love.graphics.rectangle("fill", 0, 0, self.scanlines, z)
    love.graphics.setColor(128, 64, 0)
    love.graphics.rectangle("fill", 0, z, self.scanlines, 240-z)
    for i = 1, self.scanlines do
      local size, id, offset, dist = self:raycast(i-1)
      if size then
        self.dist[i] = dist
        local bright = 0.1 + 0.9*(size/240)
        if bright > 1 then bright = 1 end
        love.graphics.setColor(255*bright, 255*bright, 255*bright)
        self.walls:draw(i-1, size, id, offset, z)
      end
    end
    self:drawSprites()
  end

  local function entCmp(a, b)
    return a.dist > b.dist
  end

  function Raycaster:drawSprites()
    for i = 1, #self.entities do
      local ent = self.entities[i]
      ent.dist = (self.x-ent.x)*(self.x-ent.x) + (self.y-ent.y)*(self.y-ent.y)
    end
    table.sort(self.entities, entCmp)
    local invDet = 1.0 / (self.planeX * self.dirY - self.dirX * self.planeY)
    for i = 1, #self.entities do
      local ent = self.entities[i]
      local sx, sy = ent.x - self.x, ent.y - self.y
      
      local transformX = invDet * ((self.dirY * sx) - (self.dirX * sy))
      local transformY = invDet * ((-self.planeY * sx) + (self.planeX * sy))
      if transformY > 0.01 then
        local screenX = math.floor((self.scanlines / 2) * (1 + transformX / transformY))
        local size = abs(math.floor(240/transformY))
        for column = math.floor(screenX-size*0.5), math.floor(screenX+size*0.5) do
          if (column>=0) and (column<self.scanlines) and (transformY<self.dist[1+column]) then
            local offset = ((screenX+size*0.5)-column)/size
            local bright = 0.1 + 0.9*(size/240)
            if bright > 1 then bright = 1 end
            love.graphics.setColor(255*bright, 255*bright, 255*bright)
            self.sprites:draw(column, size, ent.id, offset, self.planeZ)
          end
        end
      end
    end
  end

  function Raycaster:raycast(column)
    local camX = 2*column/self.scanlines-1
    local rx, ry = self.x, self.y
    local rdx, rdy = self.dirX + self.planeX*camX, self.dirY + self.planeY*camX
    local mx, my = math.floor(rx), math.floor(ry)

    local distdx = math.sqrt(1+(rdy*rdy)/(rdx*rdx))
    local distdy = math.sqrt(1+(rdx*rdx)/(rdy*rdy))
    local sidex, sidey, stepx, stepy, side
    local hit = 0

    if rdx < 0 then
      stepx = -1
      sidex = (rx-mx)*distdx
    else
      stepx = 1
      sidex = (mx + 1.0 - rx)*distdx
    end
    if rdy < 0 then
      stepy = -1
      sidey = (ry-my)*distdy
    else
      stepy = 1
      sidey = (my+1.0-ry)*distdy
    end
    local wallid = 0
    for loopcount = 1, 15 do
      if sidex < sidey then
        sidex = sidex + distdx
        mx = mx + stepx
        side = 0
      else
        sidey = sidey + distdy
        my = my + stepy
        side = 1
      end
      wallid = self.grid:get(mx, my)
      if wallid >= 1 then break end
      if loopcount >= 15 then return end
    end
    local perpWallDist
    if side == 0 then
      perpWallDist = abs((mx-rx+(1-stepx)/2)/rdx)
    else
      perpWallDist = abs((my-ry+(1-stepy)/2)/rdy)
    end
    local offset
    if side == 1 then offset = rx + ((my-ry+(1-stepy)/2)/rdy)*rdx
    else offset = ry + ((mx-rx+(1-stepx)/2)/rdx)*rdy end
    offset = offset - math.floor(offset)
    if (side==0 and rdx<0) or (side==1 and rdy>0) then
      offset = 1.0 - offset
    end
    return abs(math.floor(240/perpWallDist)), wallid, offset, perpWallDist
  end
end


function love.load()

  local walls = Texture.new("walls.png")
  local sprites = Texture.new("sprites.png")

  local grid = Grid.new(10, 10, 0, {
    { 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 },
    { 1, 0, 0, 0, 0, 0, 0, 0, 0, 1 },
    { 1, 0, 0, 0, 0, 0, 0, 1, 0, 1 },
    { 1, 0, 0, 2, 2, 0, 0, 1, 0, 1 },
    { 1, 0, 0, 2, 2, 0, 0, 1, 0, 1 },
    { 1, 0, 0, 0, 0, 0, 0, 1, 0, 1 },
    { 1, 0, 0, 0, 0, 0, 0, 1, 0, 1 },
    { 1, 0, 1, 1, 1, 1, 1, 1, 0, 1 },
    { 1, 0, 0, 0, 0, 0, 0, 0, 0, 1 },
    { 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 },
  })

  local entities = {
    { x=4.5, y=2.5, id=1, dist=0 },
  }

  local raycaster = Raycaster.new(2.5, 2.5, 0)
  raycaster:setGrid(grid, walls)
  raycaster:setEntities(entities, sprites)
  raycaster:update()

  local fps, fps_clock = 0, 0

  function love.draw()
    raycaster:draw()
    local str = ("%.1f %.1f %.1f, %i")
      :format(raycaster.x, raycaster.y, raycaster.angle, fps)
    love.graphics.setColor(255, 255, 255)
    love.graphics.print(str, 0, 0)
  end

  function love.update(dt)
    fps_clock = fps_clock + dt
    if fps_clock >= 1.0 then
      fps_clock = fps_clock - 1.0
      fps = love.timer.getFPS()
    end

    local updateRaycaster = false
    local shift = love.keyboard.isDown("lshift") or love.keyboard.isDown("rshift")
    local alt = love.keyboard.isDown("lalt") or love.keyboard.isDown("ralt")
    local movespeed = 2 * dt * (shift and 0.35 or 1)
    local turnspeed = math.pi * dt * (shift and 0.35 or 1)

    local x, y, a = raycaster.x, raycaster.y, raycaster.angle

    if love.keyboard.isDown("left") then
      if alt then
        x, y = x+math.cos(a-halfPi)*movespeed, y+math.sin(a-halfPi)*movespeed
      else
        a = a - turnspeed
        while a < 0 do a = a + twoPi end
      end
      updateRaycaster = true
    end
    if love.keyboard.isDown("right") then
      if alt then
        x, y = x+math.cos(a+halfPi)*movespeed, y+math.sin(a+halfPi)*movespeed
      else
        a = a + turnspeed
        while a > twoPi do a = a - twoPi end
      end
      updateRaycaster = true
    end
    if love.keyboard.isDown("up") then
      x, y = x+math.cos(a)*movespeed, y+math.sin(a)*movespeed
      updateRaycaster = true
    elseif love.keyboard.isDown("down") then
      x, y = x-math.cos(a)*movespeed, y-math.sin(a)*movespeed
      updateRaycaster = true
    end
    if love.keyboard.isDown("pageup") then
      raycaster.planeZ = raycaster.planeZ + 192*dt
      if raycaster.planeZ > 230 then raycaster.planeZ = 230 end
      updateRaycaster = true
    elseif love.keyboard.isDown("pagedown") then
      raycaster.planeZ = raycaster.planeZ - 192*dt
      if raycaster.planeZ < 10 then raycaster.planeZ = 10 end
      updateRaycaster = true
    end
    if love.keyboard.isDown("home") then
      raycaster.fov = raycaster.fov + math.pi*dt
      if raycaster.fov > halfPi then raycaster.fov = halfPi end
      updateRaycaster = true
    elseif love.keyboard.isDown("end") then
      raycaster.fov = raycaster.fov - math.pi*dt
      if raycaster.fov < quarterPi then raycaster.fov = quarterPi end
      updateRaycaster = true
    end

    if updateRaycaster then
      raycaster.x, raycaster.y, raycaster.angle = x, y, a
      raycaster:update()
    end
  end

  function love.keypressed(k)
    if k == "escape" then
      love.event.quit()
    elseif k == "d" then
      log = print
    end
  end
  function love.keyreleased(k)
    if k == "d" then
      log = NULLFUNC
    end
  end
end

